raise "Can't run this file directly" unless defined? $lib_dir
# Net Responsibility
#
# Copyright (C) 2007-2011 Scott Severance & Robert Rosman
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# Author:
#	Scott Severance <http://www.scottseverance.us>

require 'sqlite3'

require "#{$lib_dir}my_date"
require "#{$lib_dir}exceptions"
require "#{$lib_dir}syslog"
require "#{$lib_dir}modules/db"
require "#{$lib_dir}filter"
require "#{$lib_dir}options"

class Logger
	include DB
	
	def initialize(opts)
		raise "No logfile given in the config file" unless opts.respond_to? :logfile
		raise "No pidfile given in the config file" unless opts.respond_to? :pidfile
		logfile = opts.logfile
		daemonize = opts.daemonize
		@pid_file = opts.pidfile
		@logfile = nil
		@logfile_name = logfile
		@urlsnarf = `which urlsnarf`.chomp
		status = $?.exitstatus
		if status != 0
			case status
				when 1
					raise "Unable to locate urlsnarf."
				else
					raise "which returned an exit status of #{status}"
			end
		end
		@trapped = proc do
			runlevel = `runlevel`.chomp.slice(-1..-1)
			log_shutdown(DateTime.now, "Signal")
			begin
				File.delete @pid_file
			rescue Errno::ENOENT
				$stderr << "WARNING: No pidfile exists!\n"
			end
			Syslog.log "Received a signal. Exiting."
			raise TrapException
		end
		@reset = proc do
			log_reset DateTime.now
			Syslog.log "Resetting."
			raise ResetException
		end
		create_database(@logfile_name) #unless File.exists? @logfile_name
		set_restrictive_permissions(@logfile_name)
	end
		
	def run(opts)
		begin
			@logfile = SQLite3::Database.new @logfile_name
			last_check = scheduled_report(opts)
			warnings = [Time.at(0)]*15
			filter = Filter.new opts
			['SIGTERM','SIGHUP','SIGKILL'].each do |signal|
				trap(signal,@trapped)
			end
			trap "SIGUSR1", @reset
			log_startup DateTime.now
			begin
				urlsnarf = IO.popen("#{@urlsnarf} -i all 2> /dev/null")
				urlsnarf.each do |line|
					line.chomp!
					url = line.gsub(/^.*(GET|POST) ([^\s]+)\s.*$/, '\2')
					time_str = line.gsub(/^[^\[]*\[([^\]]*)\].*$/, '\1')
					time_str.gsub!(/\//,' ').gsub!(/(\d{4,}):/,'\1 ')
					timestamp = DateTime.parse time_str
					filtered = filter.url_match(url)
					log_url timestamp, url, filtered
					last_check = scheduled_report(opts) if Time.now > last_check + (10*60)
					instant_report(warnings, filter, opts.instant_threshold) if filtered and opts.instant_threshold > 0
				end
				raise Interrupt # We get here if urlsnarf dies prematurely
			rescue Interrupt
				log_shutdown DateTime.now, 'Interrupt'
				begin
					File.delete @pid_file
				rescue Errno::ENOENT
					$stderr << "WARNING: No pidfile exists!\n"
				end
				Syslog.log "Caught interrupt. Exiting."
			ensure
				#urlsnarf.close unless urlsnarf.closed?
				`pgrep urlsnarf`
				if $?.exitstatus == 0
					`pkill urlsnarf`
					unless $?.exitstatus == 0
						raise "urlsnarf refuses to die! Exit status: #{$?.exitstatus}"
					end
				end
			end
		ensure
			begin
				@logfile.close unless @logfile.closed?
			rescue SQLite3::BusyException => e
				#FIXME: Placeholder for exception handling
				Syslog.log_exception(
					e,
					'The following exception (SQLite3::BusyException) is caused ' +
					'by a known bug. If you know how to fix it, do tell! ' +
					'Otherwise, you can safely ignore it (I hope).',
					'daemon',
					'debug')
			end
		end
	end
	
	private
		
	def set_restrictive_permissions(*files)
		files.each do |f|
			begin
				unless File.file?(f)
					File.open(f,'w') {|file| }
				end
					File.chown(0,0,f)
					File.chmod(0600, f)
			rescue Errno::EPERM
				raise PermissionsException, "You must be root to alter these permissions."
			end
		end
	end
	 
	def scheduled_report(options = Options.new)
		if options.report_frequency > 0
			d = last_completed_report(true, 1, true)
			if d.empty?
				`#{$lib_dir}../net-responsibility-report --insidedaemon`
			else
				d = d.to_s[0..7].to_i
				`#{$lib_dir}../net-responsibility-report --insidedaemon` if DateTime.now.compact_date >= d + options.report_frequency
			end
		end
		return Time.now
	end
	
	def instant_report(warnings, filter, threshold)
		threshold.downto(1) { |i| warnings[i] = warnings[i-1] }
		warnings[0] = Time.now
		puts warnings if warnings[threshold] == nil
		if warnings[0] < warnings[threshold] + (15*60)
			count = 0
			get_warnings(true).each do |a|
				b = filter.token_match(a[:hostname] + a[:path])
				c = filter.whitelist_match(a[:hostname]) if b
				count += 1 unless c
				break if count > threshold*0.7
			end
			
			`#{$lib_dir}../net-responsibility-report --insidedaemon --instant` if count > threshold*0.7
			warnings.fill(Time.at(0))
		end
	end
end

